
/*
xpvox copyright (c) 2005 Jan Panteltje

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "xpvox.h"


pthread_t audio_thread;
pthread_attr_t *attributes;

int record_init()
{
int tmp;
FILE *pptr;
char temp[512];

if(debug_flag)
	{
	fprintf(stderr, "record_init(): arg none\n");
	}

/* set mixer for record */

if(set_mixer_flag)
	{
	sprintf(temp, "cmix rcs igain on");
	pptr = popen(temp, "r");
	pclose(pptr);

	sprintf(temp, "cmix mic 100");	
	pptr = popen(temp, "r");
	pclose(pptr);

	sprintf(temp, "cmix igain 50");
	pptr = popen(temp, "r");
	pclose(pptr);

	sprintf(temp, "cmix vol 100");
	pptr = popen(temp, "r");
	pclose(pptr);
	} /* end if set_mixer_flag */

audio = open(audio_input_device, O_RDONLY, 0);
if(audio < 0)
	{
	fprintf(stderr, "record_init(): could not open %s for read\n", audio_input_device);
		
	return 0;
	}

tmp = samplesize;
ioctl(audio, SNDCTL_DSP_SAMPLESIZE, &samplesize);
if(tmp != samplesize)
	{
	fprintf(stderr, "record_init(): unable to set the sample size\n");
	return 0;
	}

if(ioctl(audio, SNDCTL_DSP_STEREO, &dsp_stereo) == -1)
	{
	fprintf (stderr, "record_init(): unable to set mono/stereo\n");

	return 0;
	}

if(ioctl(audio, SNDCTL_DSP_SPEED, &dsp_speed) == -1)
	{
	fprintf (stderr, "record_init(): unable to set audio speed\n");
	return 0;
	}

if(debug_flag)
	{
	fprintf (stderr, "Speed %d Hz ", dsp_speed);
	if (dsp_stereo) fprintf (stderr, "(stereo)\n");
	else fprintf (stderr, "(mono)\n");
	if (samplesize != 8) 
		{
		fprintf(stderr, "record_init(): %d bits per sample\n", samplesize);
		}
    }

ioctl(audio, SNDCTL_DSP_GETBLKSIZE, &abuf_size);
if(abuf_size < 1)
	{
	fprintf(stderr, "record_init(): DSP_GETBLKSIZE failed, because ");

	perror ("GETBLKSIZE");
	return 0;
	}

if( (audiobuf = malloc (abuf_size) ) == NULL)
	{
	fprintf (stderr, "record_init(): unable to allocate input/output buffer\n");
	return 0;
	}

return 1;
} /* end function record_init */


int audio_init()
{
if(debug_flag)
	{
	fprintf(stderr, "audio_init(): arg none\n");
	}

if(! record_init() )
	{
	return 0;
	}

pthread_create(&audio_thread, NULL , &audio_routine, NULL);

return 1;
} /* end function audio_init */


void *audio_routine()
{
int i, a;
size_t l, tw;
int cm;
double da, dm;
double dmax;
int active = 0; // -Wall
int high_count, low_count;
uint8_t *pmbptr;
int bytes_per_sample;
double db = 0.0;

if(debug_flag)
	{
	fprintf(stderr, "audio_routine(): arg none\n");
	}

bytes_per_sample = samplesize / 8;

if(debug_flag)
	{
	fprintf(stderr, "abuf_size=%d audio=%d audiobuf=%p\n", abuf_size ,audio, audiobuf);
	}

dmax = 0;
high_count = 0;
low_count = 0;
while(1)
	{
	while(1)
		{
		errno = 0;
		l = read(audio, audiobuf, abuf_size);
		if (l == -1)
			{
			if(errno == EAGAIN)
				{
				fprintf(stderr, "audio_routine(): audio read(): EAGAIN\n");
				continue;
				}

			fprintf(stderr, "audio_routine(): audio read(): error return\n");
			return 0;
			}

		if(l == abuf_size) break;
		}

	/* AGC */
	if(agc_flag)
		{
/* WARNING only 2 Channels!! */

//		agc_level			/* 0.7, set in main, -3 dB */
//		agc_multiplier		/* 0-00,  set in main */

		/* test levels in audio buffer */ 
		pmbptr = audiobuf;
		for(i = 0; i < abuf_size / bytes_per_sample; i++)
      		{
			/* low byte first !!! */

			unsigned long p = 1 << 15;
			unsigned long ll;

			/* get main buffer as double */
			cm = *pmbptr + (256 * *(pmbptr + 1) );
			ll = cm;
			if(ll & p) /* sign bit */
				{
				ll = p - (ll ^ p);

				dm = -(double) ll / (double) p; 	
				}
			else /* positive */
				{
				dm = ll / ( (double) p );
				}

			/* range dm = -1 to +1 */

			/* modify amplitude */
			dm *= agc_multiplier;	/* -1 to +1 */

			/* level detect new value */			
			da = fabs(dm);			// 0-1
			if(da > agc_level)
				{
				/*scale back gain so samples fit */
				db = agc_level / da; /* always < 1.0 */
				if(agc_multiplier > 0.1) agc_multiplier *= db;

				agc_init_flag = 1;
				}

			if(! agc_init_flag)
				{
				/* at startup we start at gain 1x, fast increasing until we exceed the limit level for the first time */
				if(agc_multiplier < (double)AGC_MAX_GAIN) agc_multiplier += AGC_INIT_GAIN_INCREASE_SPEED;
				}
			else
				{
				/* slowly increase gain up to 34 dB (40dB = 100x) */
				/* 44100 samples / second, and 30 seconds, is 1323000 steps, so 50 / 323000 = .000037792 */
				/* a bit slower before we apply so much gain increase */
				if(agc_multiplier < (double)AGC_MAX_GAIN) agc_multiplier += AGC_GAIN_INCREASE_SPEED;
				}
				
			/* clip to prevent problems in 2 bytes integer */
			if(dm > 1.0) dm = 1.0;
			if(dm < -1.0) dm = -1.0;

			/* double data to output buffer */
			ll = floor(dm * p);
			if(dm < 0) ll += 2 * p;

			*pmbptr = ll % 256;
			*(pmbptr + 1) = ll / 256;

			/* increment buffer pointer */
			pmbptr += bytes_per_sample;
			} /* end for al values in audiobuf */

		if(debug_flag)
			{
			fprintf(stderr,\
			"AGC: agc_multiplier=%.2f audio_current_level=%02f audio_peak_level=%.2f agc_level=%02f\n",\
			agc_multiplier, audio_current_level, audio_peak_level, agc_level);
			}
		
		} /* end if agc_flag */
	/* end AGC */

	a = 0;
	if(output_flag)
		{
		if(output_no_silence_flag)
			{
			if(rts_is_on_flag) a = 1;
			}
		else
			{
			a = 1;
			}
		
		if(a)
			{
			tw = fwrite(audiobuf, 1, l, stdout);
			if(tw != l)
				{
				fprintf(stderr,\
				"audio_routine(): could only write %d of %d bytes to stderr, aborting.\n",\
				tw, l);

				exit(1);
				}
			} /* end if a */

		} /* end if output_flag */

	/* test levels in audio buffer */
	pmbptr = audiobuf;
	for(i = 0; i < abuf_size / bytes_per_sample; i++)
		{
		/* low byte first !!! */

		unsigned long p = 1 << 15;
		unsigned long ll;

		/* get main buffer as double */
		cm = *pmbptr + (256 * *(pmbptr + 1) );
		ll = cm;
		if(ll & p) /* sign bit */
			{
			ll = p - (ll ^ p);

			dm = -(double) ll / (double) p; 	
			}
		else /* positive */
			{
			dm = ll / ( (double) p );
			}

		/* range dm = -1 to +1 */
//		fprintf(stderr, "dm=%.10f\r", dm);
		
		da = fabs(dm);			// 0-1

		// avoid zero
		da += .00000000000001;
//		audio_current_level = 20 * log10(da / db);
		audio_current_level = 20 * log10(da / 1.0);

		if(audio_current_level > audio_peak_level)
			{
			audio_peak_level = audio_current_level;
			}

		/* speed at which audio peak meter goes to zero */ 
		if(audio_peak_level > minimum_db + .001) audio_peak_level -= .001;

		if(! active)
			{
			if(audio_current_level > threshold) if(high_count < high_samples) high_count++;
			if(high_count == high_samples)
				{
				low_count = 0;
				active = 1;

				if(debug_flag)
					{
					fprintf(stderr, "TRANSMIT\n");
					}

				if(! set_rts(serial_device, 1) )
					{
					fprintf(stderr, "xpvox: audio_routine: WARN set_rts(%s, 1) failed\n", serial_device);
					}

				if(play_lead_in_flag) start_player(LEAD_IN_SOUND);
				}
			}

		if(active)
			{
			if(audio_current_level < threshold)
				{
				if(low_count < low_samples) low_count++;
				}
			else low_count = 0;

			if(low_count == low_samples)
				{
				high_count = 0;
				active = 0;

				if(! transmission_lock_flag)
					{
					if(debug_flag)
						{
						fprintf(stderr, "RECEIVE\n");
						}
					
					if(play_lead_out_flag) start_player(LEAD_OUT_SOUND);

					if(! set_rts(serial_device, 0) )
						{
						fprintf(stderr, "xpvox: audio_routine: WARN set_rts(%s, 0) failed\n", serial_device);
						}
					}
				}
			}

		pmbptr += bytes_per_sample;
		} /* end for all samples in audio buffer */

//	usleep(1000);
	} /* end while read audio buffer */

} /* end function audio_routine */


int stop_audio()
{
if(debug_flag)
	{
	fprintf(stderr, "stop_audio(): arg none, using audio=%d\n", audio);
	}

if(audio < 0) return 0;

pthread_cancel(audio_thread);

close(audio);
free(audiobuf);

return 1;
} /* end function stop_audio */


int start_audio()
{
if(debug_flag)
	{
	fprintf(stderr,\
	"start_audio(): using audio_input_device=%s (%p)\n",\
	audio_input_device, audio_input_device); 
	}

if(! audio_init() ) return 0;

if(debug_flag)
	{
	fprintf(stderr, "start_audio(): audio=%d\n", audio);
	}

return 1;
} /* end function start_audio */

